/***** spin: spin.h *****/

/* Copyright (c) 1991-1998 by Lucent Technologies - Bell Laboratories     */
/* All Rights Reserved.  This software is for educational purposes only.  */
/* Permission is given to distribute this code provided that this intro-  */
/* ductory message is not removed and no monies are exchanged.            */
/* No guarantee is expressed or implied by the distribution of this code. */
/* Software written by Gerard J. Holzmann as part of the book:            */
/* `Design and Validation of Computer Protocols,' ISBN 0-13-539925-4,     */
/* Prentice Hall, Englewood Cliffs, NJ, 07632.                            */
/* Send bug-reports and/or questions to: gerard@research.bell-labs.com    */

/* dSpin extension written by Radu Iosif. Feedback: iosif@athena.polito.it */


#ifndef __SPIN_H__
#define __SPIN_H__


#include <stdio.h>
#include <string.h>
#include <ctype.h>

#ifndef PC
#include <memory.h>
#endif

#include "dfa.h"

#define STOP   1
#define BB     2
#define BE     4
#define FB     8
#define FE     16
#define JMP    32


typedef struct Lextok {
	unsigned short	ntyp;	/* node type */
        int     idx;            /* CHANGE: used only in dataflow analysis */
	short	ismtyp;		/* CONST derived from MTYP */
	int	val;		/* value attribute */
	int	ln;		/* line number */
	int	indstep;	/* part of d_step sequence */
	struct Symbol	*fn;	/* file name */
	struct Symbol	*sym;	/* symbol reference */
        struct Sequence *sq;	/* sequence */
        struct SeqList	*sl;	/* sequence list */
	struct Lextok	*lft, *rgt; /* children in parse tree */
} Lextok;

typedef struct Access {
	struct Symbol	*who;	/* proctype name of accessor */
	struct Symbol	*what;	/* proctype name of accessed */
	int	cnt, typ;	/* parameter nr and, e.g., 's' or 'r' */
	struct Access	*lnk;	/* linked list */
} Access;

typedef struct Symbol {
	char	*name;
        int     idx;            /* CHANGE: used only in dataflow analysis */
	int	Nid;		/* unique number for the name */
	short	type;		/* bit,short,.., chan,struct  */
	unsigned char	hidden;	/* bit-flags: 1=hide, 2=show, 128=typedef*/
  /* 
     bit-flags:
     1=hide, 2=show, 4=bit-equiv, 8=byte-equiv,
     16=formal par, 32=inline par, 64=heap-allocated
  */
        int parno;              /* number of corresponding formal parameter */
	unsigned char	colnr;	/* for use with xspin */
	int	nbits;		/* an optional width specifier */
	int	nel;		/* 1 if scalar, >1 if array   */
	int	setat;		/* last depth value changed   */

	int	*val;		/* runtime value(s), initl 0  */
	struct Lextok	**Sval;	/* values for structures and fptrs*/

  /* FIXME: bitsize is redundant; use nbits */

        unsigned offset;        /* heap/stack offset */
        unsigned bitsize;
        char     bitpos;        /* 0 .. 7 */ 

	int	xu;		/* exclusive r or w by 1 pid  */
	struct Symbol	*xup[2];/* xr or xs proctype  */
	struct Access	*access;/* e.g., senders and receives of chan */
	struct Lextok	*ini;	/* initial value, or chan-def */
	struct Lextok	*Slst;	/* template for structure if struct */
	struct Symbol	*Snm;	/* name of the defining struct */
	struct Symbol	*owner;	/* set for names of subfields in typedefs */
	struct Symbol	*context; /* 0 if global, or procname */

	struct Symbol	*next;	/* linked list */
        struct Symbol   *prev;
} Symbol;

typedef struct Value {
        int value;               /* The run-time value */
        int lval;                /* The run-time left value */
        int heap;                /* Heap flag */
} Value;

typedef struct Ordered {	/* links all names in Symbol table */ 
	struct Symbol	*entry;
	struct Ordered	*next;
        struct Ordered  *prev;
} Ordered;

typedef struct Queue {
	short	qid;		/* runtime q index */
	short	qlen;		/* nr messages stored */
	short	nslots, nflds;	/* capacity, flds/slot */
	int	setat;		/* last depth value changed */
	int	*fld_width;	/* type of each field */
	int	*contents;	/* the values stored */
	int	*stepnr;	/* depth when each msg was sent */
#ifdef __ALIGN
        int     align;          /* align flag */
#endif /* __ALIGN */
	struct Queue	*nxt;	/* linked list */
} Queue;

typedef struct Element {
	Lextok	*n;		/* defines the type & contents */

  /* CHANGE: fields added for dataflow analysis */

        struct bblock *block;   /* enclosing basic block */
        int vis;                /* visited flag */

	int	Seqno;		/* identifies this el within system */
	int	seqno;		/* identifies this el within a proc */
        int     tag;            /* FB, FE, BB, BE */
	unsigned char	status;	/* used by analyzer generator  */
        struct Sequence *fsq;   /* enclosing function sequence */
	struct SeqList	*sub;	/* subsequences, for compounds */        
	struct SeqList	*esc;	/* zero or more escape sequences */
        struct SeqList  *sbl;   /* begin sequences list */
        struct SeqList  *sel;   /* end sequences list */
	struct Element	*Nxt;	/* linked list - for global lookup */
	struct Element	*nxt;	/* linked list - program structure */
} Element;

typedef struct Sequence {
        Symbol *cnxt;           /* CHANGE: local context */
        Symbol *symtab;         /* CHANGE: sequence local symbol table */
        Symbol *stail;          /* symbol table tail (used in insertion)*/

	Element	*frst;
	Element	*last;		/* links onto continuations */
	Element *extent;	/* last element in original */

	int	maxel;		/* 1+largest id in sequence */
        int     block;          /* local scope block flag */

        unsigned long sz;       /* stack frame size */
        unsigned long bs;       /* stack frame base */
} Sequence;

typedef struct SeqList {
	Sequence	*this;	/* one sequence */
	struct SeqList	*nxt;	/* linked list  */
} SeqList;

typedef struct Label {
	Symbol	*s;
	Symbol	*c;
	Element	*e;
	int	visible;	/* label referenced in claim */
	struct Label	*nxt;
} Label;

typedef struct Lbreak {
	Symbol	*l;
	struct Lbreak	*nxt;
} Lbreak;

typedef struct Defun
{
  Symbol *fun;           /* function name   */
  int    tn;             /* ordinal */
  Lextok *prms;          /* parameter list  */
  Sequence *body;        /* executable body */
  Symbol *rt;            /* return type     */

  /* used in dataflow analysis */

  int idx;                 /* unique index number             */

  struct bb_list *bb_list; /* list of basic blocks            */
  struct bb_list *bb_dfn;  /* list of df-ordered basic blocks */

  struct Defun *nxt;     /* linked list     */
} Defun;

typedef struct RunStack {
  char *stack;
  int  size;
  int  used;
} RunStack;

typedef struct RunList {
	Symbol	*n;		/* name            */
	int	tn;		/* ordinal of type */
	int	pid;		/* process id      */
	int	priority;	/* for simulations only */
	Element	*pc;		/* current stmnt   */
        RunStack *stk;          /* CHANGE: run-time stack */
	Sequence *ps;		/* used by analyzer generator */
        Lextok	*prov;		/* provided clause */
        SeqList *rt_seq;        /* CHANGE: run-time scope */
	struct RunList	*nxt;	/* linked list */
        struct RunList  *prv;
} RunList;

typedef struct ProcList {
	Symbol	*n;		/* name       */
	Lextok	*p;		/* parameters */
	Sequence *s;		/* body       */
	Lextok	*prov;		/* provided clause */
	short	tn;		/* ordinal number */
	short	det;		/* deterministic */

  /* used in dataflow analysis */

        struct bb_list *bb_list; /* list of basic blocks            */
        struct bb_list *bb_dfn;  /* list of df-ordered basic blocks */

	struct ProcList	*nxt;	 /* linked list */
} ProcList;


typedef struct UType {
	Symbol *nm;	        /* name of the type */
	Lextok *cn;	        /* contents */
        int rt_type;            /* run-time type */

        int      ps_no;         /* list of postponed symbols */
        Symbol **ps_lst;

	struct UType *nxt;	/* linked list */
} UType;


typedef	Lextok *Lexptr;

#define YYSTYPE	Lexptr

#define ZN	(Lextok *)0
#define ZS	(Symbol *)0
#define ZE	(Element *)0

#define DONE	  1     	/* status bits of elements */
#define ATOM	  2     	/* part of an atomic chain */
#define L_ATOM	  4     	/* last element in a chain */
#define I_GLOB    8		/* inherited global ref    */
#define DONE2	 16		/* used in putcode and main*/
#define D_ATOM	 32		/* deterministic atomic    */
#define CHECK1	 64
#define CHECK2	128

#define Nhash	255    		/* slots in symbol hash-table */

#define XR	  1		/* non-shared receive-only */
#define XS	  2		/* non-shared send-only    */
#define XX	  4		/* overrides XR or XS tag  */

#define WORD      sizeof(long)

#ifdef __ALIGN
#define __align(x)                   \
do {                                 \
  (x) += (WORD - (x) % WORD) % WORD; \
} while(0)
#else
#define __align(x)
#endif /* __ALIGN */


#define BIT	  1		/* also equal to width in bits */
#define PREDEF	  2		/* predefined names: _p, _last */
#define UNSIGNED  4		/* val defines width in bits */
#define BYTE	  8		/* ditto */
#define SHORT	 16		/* ditto */
#define INT	 32		/* ditto */
#define MTYPE    64             /* message type */
#define	CHAN	128		/* not */
#define FTYPE   512             /* function reference type */		
#define STRUCT	1024		/* user defined structure */
#define POINTER 2048            /* reference type */

#define LOCAL_TAG   125
#define FUNCALL_TAG 126

#ifndef max
#define max(a,b) (((a) < (b)) ? (b) : (a))
#endif

#ifndef min
#define min(a,b) (((a) > (b)) ? (b) : (a))
#endif


/***** Global declarations *****/

extern FILE *tc, *th, *tt, *tm, *tb;

enum {ASSIGN, INITVAR, 
      GC0_SAVE, GC0_NOSV, 
      GC0_DEC_UNDO, GC0_INCR, 
      GC0_INC_UNDO, GC1_MAP,
      CH_MAP};


/***** Prototype definitions *****/

Element	*eval_sub(Element *);
Element *eval_expr(Element *, Lextok *, Value *);
int      eval(Lextok *);

Element	*get_lab(Lextok *, int);
Element	*huntele(Element *, int);
Element	*huntstart(Element *);
Element	*target(Element *);

Lextok	*do_unless(Lextok *, Lextok *);
Lextok	*expand(Lextok *, int);
Lextok	*getuname(Symbol *);
Symbol  *getusym(Symbol *);
Lextok	*mk_explicit(Lextok *, int, int);
Lextok	*nn(Lextok *, int, Lextok *, Lextok *);
Lextok	*cpnn(Lextok *, int, int, int);

Lextok	*rem_lab(Symbol *, Lextok *, Symbol *);
Lextok	*tail_add(Lextok *, Lextok *);

ProcList *ready(Symbol *, Lextok *, Sequence *, int, Lextok *);

SeqList	*seqlist(Sequence *, SeqList *);
void	 open_seq(char *, int, int, int);
Sequence *close_seq(int, int);
Sequence *check_seq(Sequence *, int);
void     enter_seq(Sequence *);
void     exit_seq();

Element *frst_el(Sequence *);
Element *last_el(Sequence *);

Symbol	*findloc(Symbol *);
Symbol	*findloc2(Symbol *);

RunStack *stack_init();
char     *stack_push(RunStack *, int);
void      stack_pop(RunStack *, int);
char     *stack_peek(RunStack *, int);

void     rt_enter(RunList *, Sequence *, int);
int      rt_exit(RunList *, int);

void      set_defun_proto(Lextok *, Lextok *, Lextok *);
void      set_defun_body(Lextok *, Lextok *);
Defun    *getdefun(Symbol *);
int       isdefun(char *);      

int       assign(Lextok *, Value *, int);

Symbol	*break_dest(void);
void    insert_symbol(Sequence *, Symbol *);
Symbol	*lookup(char *);

char	*emalloc(int);
char    *erealloc(void *, int);
void    efree(char *);

long	Rand(void);

int	any_oper(Lextok *, int);
int	any_undo(Lextok *);
int	cast_val(int, int, int);
int	checkvar(Symbol *, int);
int	Cnt_flds(Lextok *);
int	cnt_mpars(Lextok *);
int	complete_rendez(void);
int	enable(Lextok *);
int	Enabled0(Element *);
int	find_lab(Symbol *, Symbol *, int);
int	find_maxel(Symbol *);
int	getlocal(Lextok *);
int	getval(Lextok *);
int     global_assign(void);
int	has_typ(Lextok *, int);
int	ismtype(char *);
int	isproctype(char *);
int	isutype(char *);
int	Lval_struct(Lextok *, Symbol *, int);
int	main(int, char **);
int	pc_value(Lextok *);
int	proper_enabler(Lextok *);
int	putcode(FILE *, Sequence *, Element *, int, int);
int	q_is_sync(Lextok *);
int	qlen(Lextok *);
int	qfull(Lextok *);
int	qmake(Symbol *);
int	qrecv(Lextok *, int);
int	qsend(Lextok *);
int	remotelab(Lextok *);
int	remotevar(Lextok *);
int	Rval_struct(Lextok *, Symbol *);
int	setlocal(Lextok *, int);
int	setval(Lextok *, int);
int	sputtype(char *, int);
int	Sym_typ(Lextok *);
int	tl_main(int, char *[]);
int	Width_set(int *, int, Lextok *);
int	yyparse(void);
int	yywrap(void);
int	yylex(void);

void    initlocals(Sequence *);
void	add_seq(Lextok *);
void	alldone(int);
void	announce(char *);
void	chanaccess(void);
void	checkrun(Symbol *, int);
void	comment(FILE *, Lextok *, int);
void	cross_dsteps(Lextok *, Lextok *);
void	doq(Symbol *, int, RunList *);
void	dotag(char *);
void	do_var(int, Symbol *);
void	dump_struct(Symbol *, char *, RunList *);
void	dumpclaims(FILE *, int, char *);
void	dumpglobals(void);
void	dumplabels(void);
void	dumplocal(RunList *);
void	dumpsrc(int, int);
void	exit(int);
void	fatal(char *, char *);
void	fix_dest(Symbol *, Symbol *);
void	genaddproc(void);
void	genaddqueue(void);
void    genaddheap(void);
void	genheader(void);
void	genother(void);
void	gensrc(void);
void	gensvmap(void);
void	genunio(void);
void	ini_struct(Symbol *);
void	make_atomic(Sequence *, int);
void	match_trail(void);
void	nochan_manip(Lextok *, Lextok *, int);
void	non_fatal(char *, char *);
void	ntimes(FILE *, int, int, char *c[]);
void	p_talk(Element *, int);
void	pickup_inline(Symbol *, Lextok *);
void	prehint(Symbol *);
void	prep_inline(Symbol *, Lextok *);
void	pstext(int, char *);
void	pushbreak(void);
void	putname(FILE *, char *, Lextok *, int, char *, int);
void	putremote(FILE *, Lextok *, int);
void	putskip(int);
void	putsrc(int, int);
void	putstmnt(FILE *, Lextok *, int);
void	putunames();
void	rem_Seq(void);
void	runnable(ProcList *, int, int);
void    setglobals(void);
void	sched(void);
void	setaccess(Symbol *, Symbol *, int, int);
void	setmtype(Lextok *);
void	setpname(Lextok *);
void	setptype(Lextok *, int, Lextok *);
void	setuname(Lextok *);
void	setutype(Lextok *, Symbol *, Lextok *);
void	setxus(Lextok *, int);
void	Srand(unsigned);
void	start_claim(int);
void	struct_name(Lextok *, Symbol *, char *);
void	symdump(void);
void	trackchanuse(Lextok *, Lextok *, int);
void	trackvar(Lextok *, Lextok *);
void	trackrun(Lextok *);
void	trapwonly(Lextok *, char *);	/* spin.y and main.c */
char   *typ2c(Symbol *);
char   *dep2c(Symbol *);
int	typ_ck(int, int, char *);
void	undostmnt(Lextok *, int);
void	unrem_Seq(void);
void	unskip(int);
void	varcheck(Element *, Element *);
void	whoruns(int);
void	wrapup(int);
void	yyerror(char *, ...);


#endif /* __SPIN_H__ */
